#include <SPI.h>
#include <Adafruit_GFX.h>
#include <Adafruit_PCD8544.h>
#include <Fonts/Picopixel.h>
#include <Button.h>
#include <Encoder.h>

#include "logo.h"

// SCLK, DIN, D/C, SCE, RST
// 10k, 10k, 10k, 1k, 10k
Adafruit_PCD8544 display = Adafruit_PCD8544(8, 9, 19, 17, 18);

Button button1(4);
Button button2(6);
Button button3(7);
Encoder encoder(2, 3);

// General vars
unsigned long time;
int dt = 0;
boolean connected = false;
boolean dataReady = false;
boolean editing = false;

boolean blink = false;
int interval = 0;

// Encoder related vars
long position = 0;
boolean input = false;

// Serial related vars
const byte numChars = 32;
char receivedChars[numChars];
char tempChars[numChars];
char command[numChars] = {0};
boolean newData = false;

// Screen related vars
int light = 128;
int contrast = 60;
int lightTimeout = 20;

// Set variables for menu
int programLenght;
int currentProgram = 0;
char programName[numChars] = {0};
float volume = -1;

void setBacklight(int value) // Set backlight
{
  analogWrite(5, value);
  if (value != 0)
  {
    light = value;
  }
}

void setContrast(int value) // Set contrast
{
  display.setContrast(value);
  contrast = value;
}

void drawMenu() // Draw screen
{
  display.clearDisplay();
  display.setFont(&Picopixel);

  // Check for connection
  if (connected && dataReady)
  {
    // Draw top bar
    display.setCursor(0, 4);
    display.print((String)(currentProgram + 1) + "/" + (String)(programLenght));

    if (button2.read() == Button::PRESSED)
    {
      display.setCursor(73, 4);
      display.print((String)round(contrast));
    }
    else if (button3.read() == Button::PRESSED)
    {
      display.setCursor(73, 4);
      display.print((String)round(((float)light / 255) * 100));
    }
    else
    {
      display.setCursor(73, 4);
      display.print("USB");
    }
  }

  display.drawLine(0, 6, 84, 6, BLACK);
  display.setFont(NULL);

  // Check for connection
  if (connected && dataReady)
  {
    // Program name
    drawName();

    // Volume Bar
    drawBar();
  }
  else
  {
    // Blink connection message if not connected
    display.drawBitmap(0, 7, logo, 84, 16, 1);
    display.setFont(&Picopixel);

    if (!blink)
    {
      if (!connected)
      {
        display.setCursor(4, 28);
        display.print("Waiting for connection");
      }
      else if (!dataReady)
      {
        display.setCursor(15, 28);
        display.print("Waiting for data");
      }
    }

    display.setFont(NULL);
  }

  // Bottom Bar
  display.drawLine(0, 38, 84, 38, BLACK);
  if (connected && dataReady)
  {
    display.setCursor(0, 40);
    if (editing)
    {
      display.print(" -    OK    + ");
    }
    else
    {
      display.print(" <    OK    > ");
    }
  }

  display.display();
}

void drawBar() // Draw volume bar and text
{
  int pos[2] = {4, 26};

  // Set fonts and size
  display.setFont(&Picopixel);
  display.setCursor(4, pos[1] - 2);
  display.print("Volume:");

  // Center percentage
  if (volume == 0)
  {
    display.setCursor(39, pos[1] - 2);
  }
  else if (volume == 100)
  {
    display.setCursor(36, pos[1] - 2);
  }
  else
  {
    display.setCursor(37, pos[1] - 2);
  }

  if (!blink || !editing)
  {
    display.print((String)round(volume) + "%");
  }

  // Draw volume bar
  display.drawRect(pos[0], pos[1], 76, 8, BLACK);
  display.drawLine(pos[0], pos[1] + 8, pos[0], pos[1] + 9, BLACK);
  display.drawPixel(pos[0] + (76 / 4) * 1 - 1, pos[1] + 8, BLACK);
  display.drawLine(pos[0] + (76 / 4) * 2 - 1, pos[1] + 8, pos[0] + (76 / 4) * 2 - 1, pos[1] + 9, BLACK);
  display.drawPixel(pos[0] + (76 / 4) * 3 - 1, pos[1] + 8, BLACK);
  display.drawLine(pos[0] + 75, pos[1] + 8, pos[0] + 75, pos[1] + 9, BLACK);

  display.fillRect(pos[0], pos[1], round(volume / 100 * 76), 8, BLACK);
  display.setFont(NULL);
}

void drawName() // Draw program name in the middle of the display or scroll it
{
  int width = 0;
  for (size_t i = 0; i < sizeof(programName); i++)
  {
    if (programName[i] != NULL)
    {
      width = width + 6;
    }
  }

  if (width > 84)
  {
    // TODO: Make the text scroll if the name is too long
  }
  else
  {
    int x = (84 / 2) - (width / 2);
    display.setCursor(x, 10);
  }
  display.print(programName);
}

void processData() // Process serial data
{
  if (newData == true)
  {
    strcpy(tempChars, receivedChars);

    char *strtokIndx;

    strtokIndx = strtok(tempChars, ",");
    strcpy(command, strtokIndx);

    strtokIndx = strtok(NULL, ",");

    if (strcmp("HELLO", command) == 0)
    {
      connected = true;
      Serial.println("<OK>");
      Serial.println("<COUNT>");
      Serial.println("<SWITCH," + (String)currentProgram + ">");
    }

    if (strcmp("COUNT", command) == 0)
    {
      programLenght = atoi(strtokIndx);
      Serial.println("<OK>");
    }

    if (strcmp("SWITCH", command) == 0)
    {
      currentProgram = atoi(strtokIndx);
      Serial.println("<OK>");
      Serial.println("<SWITCH," + (String)currentProgram + ">");
    }

    if (strcmp("NAME", command) == 0)
    {
      strcpy(programName, strtokIndx);
      Serial.println("<OK>");
    }

    if (strcmp("GETVOL", command) == 0)
    {
      volume = atoi(strtokIndx);
      Serial.println("<OK>");
    }

    newData = false;
    dataReady = true;
  }
}

void checkSerial() // Get data from serial bus
{
  static boolean recvInProgress = false;
  static byte ndx = 0;
  char startMarker = '<';
  char endMarker = '>';
  char rc;

  while (Serial.available() > 0 && newData == false)
  {
    rc = Serial.read();

    if (recvInProgress == true)
    {
      if (rc != endMarker)
      {
        receivedChars[ndx] = rc;
        ndx++;
        if (ndx >= numChars)
        {
          ndx = numChars - 1;
        }
      }
      else
      {
        receivedChars[ndx] = '\0'; // terminate the string
        recvInProgress = false;
        ndx = 0;
        newData = true;
      }
    }

    else if (rc == startMarker)
    {
      recvInProgress = true;
    }
  }

  processData();
}

void checkInput() // Check for button input
{
  input = false;
  int value = 0;
  long newPosition = encoder.read();

  // Check if any button is pressed
  if (button1.read() == Button::PRESSED || button2.read() == Button::PRESSED || button3.read() == Button::PRESSED)
  {
    input = true;
  }

  // Encoder button
  if (button1.pressed())
  {
    editing = (editing) ? 0 : 1;
  }

  if (newPosition != position)
  {

    if (newPosition < position - 3) // Rotating anti-clockwise
    {
      value = -1;
    }
    else if (newPosition > position + 3)
    { // Rotating clockwise
      value = 1;
    }

    if (value != 0)
    {
      position = newPosition;
      input = true;

      if (button2.read() == Button::PRESSED)
      {
        // Change contrast
        if (!(contrast + value < 0) && !(contrast + value > 100))
        {
          setContrast(contrast + value);
        }
      }
      else if (button3.read() == Button::PRESSED)
      {
        // Change backlight
        if (!(light + value < 0) && !(light + value > 254))
        {
          setBacklight(light + value);
        }
      }
      else if (!editing)
      {
        // Change program
        if (currentProgram + value < 0)
        {
          currentProgram = programLenght - 1;
        }
        else if (currentProgram + value > programLenght - 1)
        {
          currentProgram = 0;
        }
        else
        {
          currentProgram = currentProgram + value;
        }

        Serial.println("<SWITCH," + (String)(currentProgram) + ">");
      }
      else
      {
        // Change volume
        if (!(volume + value < 0) && !(volume + value > 100))
        {
          Serial.println("<SETVOL," + (String)(volume + value) + ">");
        }
      }
    }
  }
}

void setup() // Init arduino
{
  // Start serial connection
  Serial.begin(115200);

  // Init button
  button1.begin();
  button2.begin();
  button3.begin();

  // Init display
  display.begin();

  setContrast(contrast);

  display.clearDisplay();
  display.display();
}

void loop() // Loop code
{
  // Check time
  dt = millis() - time;
  time = millis();

  // Check for serial data or commands
  checkSerial();

  if (programLenght == 0 || volume == -1 || strcmp("", programName) == 0)
  {
    dataReady = false;
  }

  if (!editing && connected && dataReady)
  {
    blink = false;
  }

  if (interval >= 500) // Aprox every half second, every loop takes about 25 ms
  {
    interval = 0;
    blink = (!blink) ? true : false;
    lightTimeout--;
  }
  else
  {
    interval = interval + dt;
  }

  if (input || newData)
  {
    lightTimeout = 20;
    setBacklight(light);
  }

  if (lightTimeout <= 0)
  {
    setBacklight(0);
    editing = false;
  }

  // Check if serial connection is established
  if (connected && dataReady)
  {
    drawMenu();

    // Check input
    checkInput();
  }
  else
  {
    // Draw menu with "Waiting for connection", name and volume are ignored bij drawMenu()
    drawMenu();
  }
}
