from pycaw.pycaw import AudioUtilities
from pprint import pprint
import serial
import serial.tools.list_ports
import time
import json

def main():
    print ("\nArduino Mixer by Arne van Iterson\n")
    
    # Get audio sessions
    sessions = AudioUtilities.GetAllSessions()
    print (str(len(sessions)) + " audio sessions found:")
    for session in sessions:
        print ("  " + str(session.DisplayName) + ": " + str(session.Process and session.Process.name()))

    # Get data from data.json
    with open("./res/data.json") as file:
        jsonData = json.load(file)
        print("\nArduino Device ID found:\n  " + str(jsonData['device-id']))

        print("\nCustom names found:")
        print("  System sounds -> " + str(jsonData['system-sounds']))
        for process, name in jsonData['custom-names'].items():
            print("  " + str(process) + " -> " + str(name))

    # Init serial connection
    board = serial.Serial()
    board.baudrate = 115200

    ports = serial.tools.list_ports.comports()
    if len(ports) == 0:
        print("\nNo Serial ports Available")
        exit()
    else:
        print ("\nSerial ports Available:")
        for port, desc, hwid in sorted(ports):
            print("  {}: {} [{}]".format(port, desc, hwid))
            # Check if hwid matches id found in data.json and select if so
            if jsonData["device-id"] in hwid:
                board.port = port

        # Ask for port if a match is not found
        if not (board.port):
            board.port = "COM" + input("\nSelect COM port number: ")

        # Connect to serial port
        print("Connecting to " + str(board.port))
        board.open()

        # Reset arduino and wait for it to finish startup
        print ("Waiting for arduino...")
        time.sleep(5)
        board.write("<HELLO>".encode("utf-8"))

        currentProgram = 0

        line = []
        
        newData = False
        recvInProgress = False
        startMarker = '60' # <
        endMarker = '62' # >

        timeout = 30
        t = time.time()

        while timeout != 0:
            # Keep track of time
            # TODO: Add timeouts to detect a disconnect
            dt = time.time() - t
            t = time.time()

            # Check for command markers
            for c in board.read():
                if c == 62:
                    recvInProgress = False
                    newData = True
                if recvInProgress == True:
                    line.append(chr(c))
                if c == 60:
                    recvInProgress = True
                break

            if newData == True:
                dataArray = ''.join(line).split(',')
                command = dataArray[0]
                if len(dataArray) > 1:
                    data = dataArray[1]
                
                # Process data
                # TODO: Listen for programs being added, then send this again
                if command == "COUNT":
                    board.write(("<COUNT," + str(len(sessions)) + ">").encode())
                elif command == "SWITCH":
                    currentProgram = int(data)
                    program = sessions[currentProgram]
                    displayName = None

                    volume = program.SimpleAudioVolume.GetMasterVolume()
                    
                    # Always rename system sounds because the name is too long to be handled by the Arduino
                    if program.DisplayName == "@%SystemRoot%\System32\AudioSrv.Dll,-202":
                        displayName = jsonData['system-sounds']
                    elif program.Process and program.Process.name():
                        # Check custom names from data.json
                        for process, name in jsonData['custom-names'].items():
                            if program.Process and program.Process.name() == process:
                                displayName = str(name)
                        # If no name is found check for a DisplayName
                        if displayName is None:
                            if program.DisplayName != "":
                                displayName = str(program.DisplayName)
                            else:
                                # If no DisplayName is found, use the process name
                                # TODO: Remove '.exe'
                                displayName = str(program.Process and program.Process.name())
                    
                    # Write data to arduino
                    board.write(("<NAME," + displayName + ">").encode())
                    board.write(("<GETVOL," + str(round(volume * 100)) + ">").encode())
                elif command == "SETVOL":
                    # Change volume of selected program
                    # TODO: Send volume to arduino if volume is changed some other way
                    audio = sessions[currentProgram].SimpleAudioVolume

                    audio.SetMasterVolume(float(data) / 100, None)

                    volume = audio.GetMasterVolume()

                    # Send confirmation to arduino
                    board.write(("<GETVOL," + str(round(volume * 100)) + ">").encode())

                line = []
                newData = False

if __name__ == "__main__":
    main()